#lang racket
(require games/parcheesi/play-game
         games/parcheesi/best-players
         games/parcheesi/admin
         framework
         "remote-player.rkt"
         "parse.rkt"
         racket/gui/base)

(define max-players 20)

(define always-the-same-board-and-dice #f)
(define the-port 8000)

(define (read-move-and-dice filename)
  (call-with-input-file filename
    (λ (port)
      (define color (read-line port))
      (unless (member color '("red" "green" "blue" "yellow"))
        (raise-user-error 'tournament "expected the first line to contain a color, got ~s" color))
      (define board (parse-board (rcv port)))
      (define dice (parse-dice (rcv port)))
      (set! always-the-same-board-and-dice (vector (string->symbol color)
                                                   board
                                                   dice)))))

(command-line
 #:once-each
 [("--always-the-same-board-and-dice" "-m")
  move.xml
  "\n     do not play a real game; instead repeatedly supply the same board and dice, from move.xml"
  (read-move-and-dice move.xml)]
 [("--port" "-p")
  port
  "port to run the tournament server"
  (let ([the-port (string->number port)])
    (unless (exact-nonnegative-integer? the-port)
      (raise-user-error
       'tournament "expected a port number on the command line, found ~a" 
       port)))])

;; name : string
;; games : number  -- number of games played
;; obj : (is-a?c player<%>)
(define-struct player (name [games #:mutable] obj) #:transparent)
  
(define player-frame (new frame% 
                          (label "Parcheesi Tournament")
                          (width 200)
                          (height 500)))
(define hp (new horizontal-panel% [parent player-frame]))
(define player-pane (new vertical-pane% [parent hp]))
(define lots-pane (new vertical-pane% [parent hp]))
(define list-box (new list-box%
                      (label #f)
                      (choices '())
                      (parent player-pane)
                      (callback void)))
  
(define start-button (new button%
                          (label (if always-the-same-board-and-dice
                                     "Send Fixed Board/Dice To All Players"
                                     "Start Tournament"))
                          (parent player-pane)
                          (callback
                           (lambda (x y)
                             (disable-gui)
                             (channel-put start-tournament-chan
                                          always-the-same-board-and-dice)))))
(define lots-button (new button%
                         [label "Run Many Games"]
                         [parent player-pane]
                         [callback
                          (λ (x y)
                            (disable-gui)
                            (channel-put lots-games-chan (void)))]))

(define (disable-gui)
  (send start-button enable #f)
  (send lots-button enable #f))
(define (enable-gui)
  (send start-button enable #t)
  (send lots-button enable #t))

(define (refresh-win-percentages rounds-played win-percentages)
  (when (null? (send lots-pane get-children))
    (define t (new text:standard-style-list%))
    (new message% [parent lots-pane] [label "Win Percentages"])
    (define ec (new editor-canvas% [parent lots-pane] [editor t]))
    (send ec min-width 400))
  (define t (send (list-ref (send lots-pane get-children) 1) get-editor))
  (send t begin-edit-sequence)
  (send t erase)
  (define p (open-output-text-editor t))
  (define sorted-win-percentages
    (sort win-percentages > #:key (λ (x)
                                    (define n (cadr x))
                                    (cond
                                      [(number? n) n]
                                      [else -inf.0]))))
  (for ([win-percentage (in-list sorted-win-percentages)])
    (define wp-n (list-ref win-percentage 1))
    (fprintf p "~a: ~a\n"
             (cond
               [(not (number? wp-n)) "illegal move"]
               [(= wp-n +inf.0) "∞    "]
               [else (~r #:precision '(= 4) wp-n)])
             (list-ref win-percentage 0)))
  (fprintf p "\n~a rounds played" rounds-played)
  (send t end-edit-sequence))

(define (add-player player)
  (send list-box append (player-name player) player))
  
;; connections-chan : (channel (cons input-port output-port))
(define connections-chan (make-channel))
  
;; start-tournament-chan : (channel (void))
(define start-tournament-chan (make-channel))

;; lots-games-chan : (channel (void))
(define lots-games-chan (make-channel))

(void
 (let ([listener (tcp-listen the-port 12 #t)])
   (thread
    (lambda ()
      (let loop ()
        (let-values ([(in out) (tcp-accept listener)])
          (channel-put connections-chan (cons in out))
          (loop)))))))
  
;; show frame after starting listener
(send player-frame show #t)

(void
 (thread
  (lambda ()
    (let loop ([connections '()])
      (sync
       (handle-evt
        connections-chan
        (lambda (pr)
          (cond
            [(< (length connections) max-players)
             (let* ([in (car pr)]
                    [out (cdr pr)]
                    [name (get-name in out)])
               (cond
                 [name
                  (let ([player (make-player name
                                             0 
                                             (new remote-player% 
                                                  [in in]
                                                  [out out]
                                                  [override-name name]))])
                    (queue-callback
                     (lambda ()
                       (add-player player)))
                    (loop (cons player connections)))]
                 [else (loop connections)]))]
            [else
             (printf "too many players, dropping player\n")
             (loop connections)])))
       (handle-evt
        start-tournament-chan
        (lambda (always-the-same-board-and-dice)
          (printf "starting tournament\n")
          (start-tournament always-the-same-board-and-dice connections)
          (queue-callback (λ () (enable-gui)))
          (loop connections)))
       (handle-evt
        lots-games-chan
        (lambda (_)
          (printf "running lots of games\n")
          (run-lots-of-games connections)
          (queue-callback (λ () (enable-gui)))
          (loop connections))))))))
    
(define names-table (make-hash))
(define (get-name in out)
  (with-handlers ([exn:fail?
                   (lambda (x) 
                     (printf "error getting name ~s\n" x)
                     #f)])
    (snd `(start-game "red") out)
    (let ([network-name
           (match (rcv in)
             [`(name () ,(? string? name)) name]
             [x (error 'get-name "bad name ~s" x)])])
      (let loop ([n #f])
        (let ([name-candidate (if n
                                  (format "~a #~a" network-name n)
                                  network-name)])
          (let ([bound? (hash-ref names-table name-candidate (lambda () #f))])
            (cond
              [bound? (loop (if n (+ n 1) 1))]
              [else
               (hash-set! names-table name-candidate #t)
               name-candidate])))))))
  
(define (start-tournament always-the-same-board-and-dice players)
  (cond
    [always-the-same-board-and-dice
     (for ([player (in-list players)])
       (printf "sending fixed move to ~a\n" (player-name player))
       (send (player-obj player) start-game (vector-ref always-the-same-board-and-dice 0))
       (define response
         (send (player-obj player) do-move
               (vector-ref always-the-same-board-and-dice 1)
               (vector-ref always-the-same-board-and-dice 2)))
       (printf "got response:\n")
       (snd (unparse-moves response) (current-output-port)))]
    [else
     (let loop ([winners '()]
                [good-players players]
                [round 1])
       (cond
         [(= (length winners) 4)
          (printf "winner ~s\n"
                  (channel-get (play-game/thread
                                winners
                                #:name "Last Round")))]
         [(<= (length good-players) (- 4 (length winners)))
          (loop (pad-to-4-players (append winners good-players))
                '()
                round)]
         [else
          (let* ([chans (start-many-games good-players (- 4 (length winners)) round)]
                 [local-winners/cheaters (map (lambda (c) (channel-get c)) chans)]
                 [local-winners (map (lambda (x) (name->player x players))
                                     (filter (lambda (x) x) 
                                             (map car local-winners/cheaters)))]
                 [cheaters (map (lambda (x) (name->player x players))
                                (apply append (map cadr local-winners/cheaters)))])
            (loop (append local-winners winners)
                  (remove-many local-winners (remove-many cheaters good-players))
                  (+ round 1)))]))]))

(define (run-lots-of-games _players)
  (define all-player-names
    (for/set ([player (in-list _players)])
      (player-name player)))
  (define wins (make-hash))
  (define games (make-hash))
  
  (let loop ([rounds-played 0]
             [_players _players])
    (define players
      (cond
        [(<= (length _players) 4)
         (pad-to-4-players _players)]
        [else _players]))
    (define this-round-players (shuffle players))
    (define game-threads
      (for/list ([i (in-range (floor (/ (length this-round-players) 4)))])
        (define this-game-players (take (drop this-round-players (* i 4)) 4))
        (for ([this-game-player (in-list this-game-players)])
          (hash-set! games (player-name this-game-player)
                     (+ 1 (hash-ref games (player-name this-game-player) 0))))
        (play-game/thread this-game-players #:gui? #f)))
    (define game-results (for/list ([thd (in-list game-threads)])
                           (sync thd)))
    (for ([game-result (in-list game-results)])
      (define-values (winner cheaters) (apply values game-result))
      (let ([last-wins (hash-ref wins winner 0)])
        (when (number? last-wins)
          (hash-set! wins winner (+ last-wins 1))))
      (for ([cheater (in-list cheaters)])
        (printf "cheated! ~s\n" cheater)
        (hash-set! wins cheater #f)
        (set! this-round-players (remove-cheater cheater this-round-players))))

    
    ;; show status
    (let ()
      (define win-percentages
        (for/list ([player-name (in-set all-player-names)])
          (define illegal-move? (not (hash-ref wins player-name 0)))
          (cond
            [illegal-move?
             (list player-name #f)]
            [else
             (define this-player-wins (hash-ref wins player-name 0))
             (define this-player-games (hash-ref games player-name 0))
             (define win-percentage (if (= 0 this-player-games)
                                        +inf.0
                                        (/ this-player-wins
                                           this-player-games)))
             (list player-name win-percentage)])))
      (queue-callback
       (λ ()
         (refresh-win-percentages (+ 1 rounds-played) win-percentages))))
    
    (loop (+ rounds-played 1)
          this-round-players)))

(define (remove-cheater name players)
  (for/list ([player (in-list players)]
             #:unless (equal? name (player-name player)))
    player))
    

(define (pad-to-4-players players)
  (append players (make-standin-players (- 4 (length players)))))
    
(define (make-standin-players n)
  (let loop ([n n])
    (cond
      [(zero? n) '()]
      [else 
       (let ([name (format "Standin Random Player ~a" n)])
         (cons (make-player name 0 (new random-player% [name name]))
               (loop (- n 1))))])))
  
;; start-many-games : (listof players) ? natural -> (listof chan)
;; starts as many games as possible in players
;; effect: updates the player-games count
;; returns the channels that tell the results of the games
(define (start-many-games players max-games round)
  (let loop ([players players]
             [max-games max-games])
    (cond
      [(zero? max-games) '()]
      [else
       (let ([to-play (pick-four-players players)])
         (if to-play
             (begin
               (for-each (lambda (player) (set-player-games! player (+ (player-games player) 1)))
                         to-play)
               (cons (play-game/thread to-play #:name (~a "Round " round))
                     (loop (remove-many to-play players) (- max-games 1))))
             '()))])))
  
(define (name->player name players)
  (cond
    [(not name) #f]
    [else (let loop ([players players])
            (cond
              [(null? players) #f]
              [else (if (equal? name (player-name (car players)))
                        (car players)
                        (loop (cdr players)))]))]))
              
;; pick-four-players : (listof players) -> (union #f (list player player player player))
(define (pick-four-players players)
  (cond
    [(< (length players) 4) #f]
    [else
     (let loop ([picked-players '()]
                [unpicked-players players])
       (let* ([min-count (apply min (map player-games unpicked-players))]
              [min-players (filter (lambda (x) (= (player-games x) min-count))
                                   players)])
         (cond
           [(< (+ (length min-players) (length picked-players)) 4)
            (loop (append min-players picked-players)
                  (remove-many min-players unpicked-players))]
           [else (append picked-players 
                         (randomly-select (- 4 (length picked-players)) min-players))])))]))
  
;; remove-many : list-of-X list-of-X -> list-of-X
;; removes each element of the first list from the second list and returns that
(define (remove-many to-remove main-list) (foldl remove main-list to-remove))
  
(define (randomly-select n lst)
  (cond
    [(zero? n) null]
    [else 
     (let ([selected (list-ref lst (random (length lst)))])
       (cons selected (randomly-select (- n 1) (remove selected lst))))]))
  
(define (play-game/thread players #:gui? [gui? #t] #:name [name #f])
  (define c (make-channel))
  (thread
   (lambda ()
     (define player-objs (map player-obj players))
     (channel-put 
      c
      (cond
        [gui? (play-game player-objs #:name name)]
        [else (play-game/no-gui player-objs)]))))
  c)

(define (play-game/no-gui players)
  (define game (new game%))
  (for-each (lambda (player) (send game register player)) players)
  (send game start))

(void
 (thread
  (λ ()
    (define lr (make-log-receiver (current-logger) 'info 'parcheesi-bad-move-information))
    (let loop ()
      (sync (handle-evt
             lr
             (λ (x)
               (define message (vector-ref x 1))
               (define color/board/dice/moves (vector-ref x 2))
               (printf "============================================================\n")
               (print-counterexample (vector-ref color/board/dice/moves 0)
                                     (vector-ref color/board/dice/moves 1)
                                     (vector-ref color/board/dice/moves 2)
                                     (vector-ref color/board/dice/moves 3))
               (printf "============================================================\n")
               (loop))))))))

(define (print-counterexample color board dice moves)
  (printf "~a\n\n" color)
  (snd (unparse-board board) (current-output-port))
  (newline)
  (snd (unparse-dice dice) (current-output-port))
  (newline)
  (snd (unparse-moves moves) (current-output-port)))
  
(module+ test
  (require rackunit)

  (define (all-ns players ns)
    (for/and ([x (in-list players)])
      (member (player-games x) ns)))

  (check-not-false (all-ns (pick-four-players (list (make-player "a" 0 #f)
                                               (make-player "b" 0 #f)
                                               (make-player "c" 0 #f)
                                               (make-player "d" 0 #f)
                                               (make-player "e" 0 #f)
                                               (make-player "f" 0 #f)
                                               (make-player "g" 0 #f)
                                               (make-player "h" 0 #f)
                                               (make-player "i" 0 #f)
                                               (make-player "j" 0 #f)))
                      (list 0)))

  (check-not-false (all-ns (pick-four-players (list (make-player "a" 0 #f)
                                               (make-player "b" 1 #f)
                                               (make-player "c" 0 #f)
                                               (make-player "d" 1 #f)
                                               (make-player "e" 0 #f)
                                               (make-player "f" 1 #f)
                                               (make-player "g" 1 #f)
                                               (make-player "h" 0 #f)
                                               (make-player "i" 0 #f)
                                               (make-player "j" 0 #f)))
                      (list 0)))

  (check-not-false (all-ns (pick-four-players (list (make-player "a" 1 #f)
                                               (make-player "b" 1 #f)
                                               (make-player "c" 1 #f)
                                               (make-player "d" 0 #f)
                                               (make-player "e" 1 #f)
                                               (make-player "f" 0 #f)
                                               (make-player "g" 1 #f)
                                               (make-player "h" 1 #f)
                                               (make-player "i" 0 #f)
                                               (make-player "j" 0 #f)))
                      (list 0)))

  (check-not-false
   (all-ns (pick-four-players (list (make-player "a" 1 #f)
                                    (make-player "b" 1 #f)
                                    (make-player "c" 1 #f)
                                    (make-player "d" 1 #f)
                                    (make-player "e" 1 #f)
                                    (make-player "f" 1 #f)
                                    (make-player "g" 1 #f)
                                    (make-player "h" 1 #f)
                                    (make-player "i" 0 #f)
                                    (make-player "j" 0 #f)))
           (list 0 1)))

  (check-false (pick-four-players (list (make-player "a" 0 #f)
                                        (make-player "b" 0 #f)
                                        (make-player "c" 0 #f)))))
